import { ChangeDetectorRef, NgZone, OnDestroy, PipeTransform } from '@angular/core';
import { RxStrategyProvider, RxStrategyNames } from '@rx-angular/cdk';
import { NextObserver, Observable, ObservableInput } from 'rxjs';
/**
 * @Pipe PushPipe
 *
 * @description
 *
 * The push pipe serves as a drop-in replacement for angulars built-in async pipe.
 * Just like the *rxLet Directive, it leverages a
 * [RenderStrategy](https://github.com/rx-angular/rx-angular/blob/master/libs/cdk/docs/render-strategies/README.md) under the hood which takes care of optimizing the ChangeDetection of your
 * component.
 * The rendering behavior can be configured per PushPipe instance using either a strategy name or provide a
 * `RxComponentInput` config.
 *
 * Usage in the template
 *
 * ```html
 * <hero-search [term]="searchTerm$ | push"> </hero-search>
 * <hero-list-component [heroes]="heroes$ | push"> </hero-list-component>
 * ```
 *
 * Using different strategies
 *
 * ```html
 * <hero-search [term]="searchTerm$ | push: 'immediate'"> </hero-search>
 * <hero-list-component [heroes]="heroes$ | push: 'normal'"> </hero-list-component>
 * ```
 *
 * Provide a config object
 *
 * ```html
 * <hero-search [term]="searchTerm$ | push: { strategy: 'immediate' }"> </hero-search>
 * <hero-list-component [heroes]="heroes$ | push: { strategy: 'normal' }"> </hero-list-component>
 * ```
 *
 * Other Features:
 *
 * - lazy rendering (see
 *  [LetDirective](https://github.com/rx-angular/rx-angular/tree/master/libs/template/docs/api/let-directive.md))
 * - Take observables or promises, retrieve their values and render the value to the template
 * - a unified/structured way of handling null, undefined or error
 * - distinct same values in a row skip not needed re-renderings
 *
 * @usageNotes
 *
 * ```html
 * {{observable$ | push}}
 * <ng-container *ngIf="observable$ | push as o">{{o}}</ng-container>
 * <component [value]="observable$ | push"></component>
 * ```
 *
 * @publicApi
 */
export declare class PushPipe<S extends string = string> implements PipeTransform, OnDestroy {
    private strategyProvider;
    private cdRef;
    private ngZone;
    /**
     * @internal
     * This is typed as `any` because the type cannot be inferred
     * without a class-level generic argument, which was removed to
     * fix https://github.com/rx-angular/rx-angular/pull/684
     */
    private renderedValue;
    /** @internal */
    private readonly subscription;
    /** @internal */
    private readonly templateObserver;
    /** @internal */
    private readonly strategyHandler;
    /** @internal */
    private patchZone;
    /** @internal */
    private _renderCallback;
    constructor(strategyProvider: RxStrategyProvider, cdRef: ChangeDetectorRef, ngZone: NgZone);
    transform<U>(potentialObservable: null, config?: RxStrategyNames<S> | Observable<RxStrategyNames<S>>, renderCallback?: NextObserver<U>): null;
    transform<U>(potentialObservable: undefined, config?: RxStrategyNames<S> | Observable<RxStrategyNames<S>>, renderCallback?: NextObserver<U>): undefined;
    transform<U>(potentialObservable: ObservableInput<U>, config?: RxStrategyNames<S> | Observable<RxStrategyNames<S>>, renderCallback?: NextObserver<U>): U;
    transform<U>(potentialObservable: ObservableInput<U>, config?: PushInput<U, S>): U;
    /** @internal */
    ngOnDestroy(): void;
    /** @internal */
    private setPatchZone;
    /** @internal */
    private handleChangeDetection;
}
interface PushInput<T, S> {
    strategy?: RxStrategyNames<S> | Observable<RxStrategyNames<S>>;
    renderCallback?: NextObserver<T>;
    patchZone?: boolean;
}
export {};
